﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ModbusMon
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            comPortOpen = false;
            noReceiveTick = NoReceiveTickMax;
            txbuff = new byte[TxBuffSize];
            txtextbuff = new byte[16];
            rcvLineBuffer = new byte[rcvBuffSize];
            rcvSize = 0;
            //reqVal = REQ_VALUE.NON;
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            buttonCloseCOM.Enabled = false;
            buttonSetCOM.Enabled = true;
            buttonOpenCOM.Enabled = true;
            //comPortName = "COM1";
            comPortName = Properties.Settings.Default.COMPort;
            comBitRate = Properties.Settings.Default.BitRate;

            asciiMode = Properties.Settings.Default.Ascii_Mode;
            radioASCII.Checked = asciiMode;
            radioRTU.Checked = !asciiMode;
            timer1.Enabled = true;
        }


        private bool comPortOpen;
        private int noReceiveTick;
        
        private bool asciiMode;

        private const int NoReceiveTickThresh   = 5;
        private const int NoReceiveTickMax      = 10;

        private string comPortName;
        private int comBitRate;

        private int updateErroCheckTick;


        private byte[] txbuff;
        private byte[] txtextbuff;
        private const int TxBuffSize = 32;

        private byte[] rcvLineBuffer;
        private const int rcvBuffSize = 1024;
        private int rcvSize;


        //-----------------------
        uint querySlaveID;
        uint queryFunctionCode;
        uint queryStartAddress;
        uint queryQuantity;

 
        private void timer1_Tick(object sender, EventArgs e)
        {
            updateErroCheckTick++;
            if (updateErroCheckTick == 20)
            {
                updateErrorCheck();
                updateErroCheckTick = 0;
            }

            if (comPortOpen)
            {
                int n = serialPort1.BytesToRead;
                if (n > 0)
                {
                    noReceiveTick = 0;
                    for(int i = 0; i < n; i++)
                    {
                        byte c = (byte)(serialPort1.ReadByte());
                        append_to_RCV_box(c);

                        if (rcvSize < rcvBuffSize) {
                            rcvLineBuffer[rcvSize++] = c;
                        }

                    }
                }
                else
                {
                    if (noReceiveTick < NoReceiveTickMax)
                    {
                        noReceiveTick++;
                        if (noReceiveTick == NoReceiveTickThresh)
                        {
                            crlf_RCV_box();
                            ReadRcvLineBuffer();
                        }
                    }

                }
            }
        }
        
        private void  append_to_RCV_box(byte b)
        {
            string str = b.ToString("X2")+" ";
            textBox1.AppendText(str);
        }

        private void crlf_RCV_box()
        {
            textBox1.AppendText("\r\n");
        }


        private void buttonOpenCOM_Click(object sender, EventArgs e)
        {
            serialPort1.PortName = comPortName;
            serialPort1.BaudRate = comBitRate;
            try
            {
                serialPort1.Open();
                comPortOpen = true;
             }
            catch
            {
                comPortOpen = false;
            }

            if (!comPortOpen)
            {
                timer1.Enabled = false;
                buttonCloseCOM.Enabled = false;
                buttonSetCOM.Enabled = true;
                MessageBox.Show("COM PORT open error");
            }
            else
            {
                timer1.Enabled = true;
                buttonCloseCOM.Enabled = true;
                buttonSetCOM.Enabled = false;
                buttonOpenCOM.Enabled = false;
            }
        }

        private void closeCom()
        {
            if (comPortOpen != false)
            {
                comPortOpen = false;
                serialPort1.Close();
                buttonCloseCOM.Enabled = false;
                buttonSetCOM.Enabled = true;
                buttonOpenCOM.Enabled = true;
            }
        }

        private void buttonCloseCOM_Click(object sender, EventArgs e)
        {
            closeCom();
        }

        private void radioASCII_CheckedChanged(object sender, EventArgs e)
        {
            asciiMode = radioASCII.Checked;
            Properties.Settings.Default.Ascii_Mode = asciiMode;
            Properties.Settings.Default.Save();
        }

        private void buttonSetCOM_Click(object sender, EventArgs e)
        {
            SetComForm form = new SetComForm();
            form.PortName = comPortName;
            form.Bitrate = comBitRate;
            DialogResult r = form.ShowDialog();
            if(r == DialogResult.OK)
            {
                comPortName = form.PortName;
                comBitRate = form.Bitrate;
                Properties.Settings.Default.COMPort = comPortName;
                Properties.Settings.Default.BitRate= comBitRate;
                Properties.Settings.Default.Save();
            }
            form.Dispose();
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            closeCom();
        }


        private int strToN(string str)
        {
            int n;
            if (str.Length == 0) return 0;
            //10進数でパースできなかったら16進数とみなす。
            if (!Int32.TryParse(str, out n)){
                try
                {
                    n = Convert.ToInt32(str, 16);
                }
                catch
                {
                    n = 0;
                }
            }
            return n;

        }


        private void get_QUERY_value()
        {
            querySlaveID = (uint)strToN(textBoxSlave.Text);
            queryFunctionCode = (uint)strToN(textBoxFunction.Text);
            queryStartAddress = (uint)strToN(textBoxAddress.Text);
            queryQuantity = (uint)strToN(textBoxQuantity.Text);
            txbuff[0] = (byte)(querySlaveID & 0xFF);
            txbuff[1] = (byte)(queryFunctionCode & 0xFF);
            txbuff[2] = (byte)((queryStartAddress >> 8) & 0xFF);
            txbuff[3] = (byte)(queryStartAddress & 0xFF);
            txbuff[4] = (byte)((queryQuantity >> 8) & 0xFF);
            txbuff[5] = (byte)((queryQuantity & 0xFF));


        }
        

        private uint get_crc(byte[]buff,int len)
        {
            uint crc_val = 0xFFFF;
            
            for(int i=0;i<len; i++)
            {
                crc_val ^= buff[i];
                for(int j=0; j < 8; j++)
                {
                    bool carry = ((crc_val & 1) != 0);
                    crc_val >>= 1;
                    if (carry)
                    {
                        crc_val ^= 0xA001;
                    }

                }
            }
            return crc_val;
        }

        private uint set_query_err_check()
        {
            uint n;
            if (asciiMode)
            {
                n = 0;
                for (int i = 0; i < 6; i++)
                {
                    n += txbuff[i];
                }
                n &= 0xFF;
                n ^= 0xFF;
                n++;
                n &= 0xFF;
                txbuff[6] = (byte)n;
            }
            else
            {
                n = get_crc(txbuff, 6);
                txbuff[6] = (byte)(n & 0xFF);
                txbuff[7] = (byte)((n >> 8) & 0xFF);

            }
            return n;
        }

        private void updateErrorCheck()
        {
            get_QUERY_value();
            uint n = set_query_err_check();
            textBoxErrCheck.Text = n.ToString("X");
        }


        private void sendQuery()
        {
            if (!comPortOpen) return;
            set_query_err_check();
            if (asciiMode) {
                StringBuilder sb = new StringBuilder();
                sb.Append(':');
                for (int i = 0; i < 7; i++) {
                    sb.Append(((uint)(txbuff[i])).ToString("X2"));
                }
                sb.Append("\r\n");
                string s = sb.ToString();
                serialPort1.Write(s);
            }
            else {
                serialPort1.Write(txbuff, 0, 8);
            }

        }


        private void buttonSend_Click(object sender, EventArgs e)
        {
            
            get_QUERY_value();
            sendQuery();
        }

        private void buttonClear_Click(object sender, EventArgs e)
        {
            textBox1.Clear();
        }

 
       

        private void ReadRcvLineBuffer()
        {
            byte[] binbuff;
            if (asciiMode) {
                binbuff = new byte[1];
                rcvAscii2Bin(ref binbuff);
            }
            else {
                if (rcvSize < 3) return;
                binbuff = new byte[rcvSize];
                for(int i = 0; i < rcvSize; i++) {
                    binbuff[i] = rcvLineBuffer[i];
                }
            }
            rcvSize = 0;
        }




        //private bool rcvAscii2Bin(byte[] responseAscii, int responref byte[] binOut)
        private bool rcvAscii2Bin(ref byte[] binOut)
        {
            int len = rcvSize;
            if (len < 7) { //先頭のコロンと LCR2文字と、CR,LFを除外した上で偶数個のデータがあるはず
                return false;
            }
            int n = len - 5;
            if ((n & 1) != 0) return false;
            n /= 2;
            if (rcvLineBuffer[0] != 0x3A) return false;
            binOut = new byte[n];
            int s = 1;
            int d = 0;
            for (; ; ) {
                byte b = 0;
                string str = string.Format("{0}{1}", (char)(rcvLineBuffer[s]), (char)(rcvLineBuffer[s + 1]));
                try {
                    b = Convert.ToByte(str, 16);
                }
                catch {
                    b = 0;
                }
                binOut[d] = b;
                d++; s += 2;
                if (d == n) break;
            }
            return true;
        }


        protected  byte highByte(uint wd)
        {
            return (byte)((((uint)wd) & 0xff00) >> 8);
        }

        protected byte lowByte(uint wd)
        {
            return (byte)(((uint)wd) & 0xff);
        }

        private void radioRTU_CheckedChanged(object sender, EventArgs e)
        {
 
        }

        private void buttonAbout_Click(object sender, EventArgs e)
        {
            AboutBox dlg = new AboutBox();
            dlg.ShowDialog();
            dlg.Dispose();
        }
    }
}
